<?php

namespace Drupal\skilling\Controller;

use Drupal\block\Entity\Block;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\AlertCommand;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Path\AliasManager;
use Drupal\Core\Theme\ThemeManager;
use Drupal\node\Entity\Node;
use Drupal\skilling\Ajax\SkillingBookTreeCommand;
use Drupal\skilling\SkillingConstants;
use Drupal\skilling\SkillingCurrentUser;
use Drupal\user\UserDataInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\skilling\Utilities as SkillingUtilities;
use Drupal\book\BookManager;
use Drupal\Core\Block\BlockManager;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\RedirectResponse;

/**
 * Class ContentNavController.
 */
class ContentNavController extends ControllerBase {

  /**
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
  * @var SkillingUtilities
  */
  protected $skillingUtilities;

  /** @var BookManager */
  protected $bookManager;

  /** @var BlockManager */
  protected $blockManager;

  /** @var ThemeManager  */
  protected $themeManager;

  /** @var SkillingCurrentUser */
  protected $currentUser;

  /**
   * @var UserDataInterface
   */
  protected $userData;

  /**
   * The alias manager service.
   *
   * @var \Drupal\Core\Path\AliasManager
   */
  protected $aliasManager;

  /**
   * Constructs a new ContentNavController object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   * @param \Drupal\skilling\Utilities $skilling_utilities
   * @param \Drupal\book\BookManager $bookManager
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    SkillingUtilities $skilling_utilities,
    BookManager $bookManager,
    BlockManager $blockManager,
    ThemeManager $themeManager,
    SkillingCurrentUser $currentUser,
    UserDataInterface $userData,
    AliasManager $aliasManager
) {
    $this->entityTypeManager = $entity_type_manager;
    $this->skillingUtilities = $skilling_utilities;
    $this->bookManager = $bookManager;
    $this->blockManager = $blockManager;
    $this->themeManager = $themeManager;
    $this->currentUser = $currentUser;
    $this->userData = $userData;
    $this->aliasManager = $aliasManager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    /** @noinspection PhpParamsInspection */
    return new static(
      $container->get('entity_type.manager'),
      $container->get('skilling.utilities'),
      $container->get('book.manager'),
      $container->get('plugin.manager.block'),
      $container->get('theme.manager'),
      $container->get('skilling.skilling_current_user'),
      $container->get('user.data'),
      $container->get('path.alias_manager')
    );
  }

  /**
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   */
  public function loadBookNavsAjax() {
    // Find instances of Skilling book nav block in the current theme.
    $themeName = $this->themeManager->getActiveTheme()->getName();
    $blockIds = $this->entityTypeManager->getStorage('block')
      ->getQuery()
      ->condition('plugin', 'skilling_book_nav_block')
      ->condition('theme', $themeName)
      ->execute();
    // Extract the bids for each.
    $bids = [];
    foreach ($blockIds as $blockId) {
      $block = Block::load($blockId);
      if ($block) {
        $settings = $block->get('settings');
        if ($settings) {
          if (isset($settings['book_to_show'])) {
            $bid = $settings['book_to_show'];
            // For students and anons...
            $anon = $this->currentUser->isAnonymous();
            $student = $this->currentUser->isStudent();
            if ($anon || $student) {
              // ... are they allowed to see design pages?
              $config = $this->config('skilling.settings');
              $seeDesignPages = $config->get('design_pages.students_anons_view');
              if (!$seeDesignPages) {
                // No.
                // Is it a design page book? Check the node that is the book root.
                /** @var Node $root */
                $root = $this->entityTypeManager->getStorage('node')
                  ->load($bid);
                if ($root) {
                  if ($root->bundle() === SkillingConstants::DESIGN_PAGE_CONTENT_TYPE) {
                    // It's a design page book.
                    // Jump to end of loop, w/o storing bid.
                    continue;
                  }
                } //End there is a root node.
              } //End can't see design pages.
            } //User is anon or student.
            $bids[] = $settings['book_to_show'];
          }
        }
      }
    } // End for.
    // Create AJAX Response object for each book.
//    $response = new AjaxResponse();
    $result = [];
    foreach ($bids as $bid) {
      // Get the title of the book.
      /** @var Node $book */
      $book = $this->entityTypeManager->getStorage('node')->load($bid);
      $title = $book->getTitle();
      // Is there a cache for this book?
      $bookTree = \Drupal::cache()->get('book' . $bid);
      if (!$bookTree) {
        // Generate tree for a book.
        $bookTree = $this->getBookTree($bid);
        \Drupal::cache()->set('book' . $bid, $bookTree);
      }
      $result[] = [
        'bid' => $bid,
        'title' => $title,
        'tree' => $bookTree,
      ];
//      $response->addCommand(new SkillingBookTreeCommand($bid, $title, $bookTree));
    } // End foreach.
    return new JsonResponse($result);
  }

  public function getBookTree($bid) {
    $bookTree = $this->bookManager->bookTreeAllData($bid);
    // Make a nested array.
    $bookTreeExtract = $this->makeBookTreeExtract(current($bookTree));
    return $bookTreeExtract;
  }

  protected function makeBookTreeExtract(array $subtreeRoot) {
    $nid = $subtreeRoot['link']['nid'];
    $alias = $this->aliasManager->getAliasByPath('/node/' . $nid);
    $alias = $this->skillingUtilities->removeSlashFromStart($alias);
    $extract = [
      'key' => $subtreeRoot['link']['nid'],
      'alias' => $alias,
      'title' => $subtreeRoot['link']['title'],
      'children' => [],
    ];
    if (isset($subtreeRoot['below'])) {
      foreach ($subtreeRoot['below'] as $child) {
        $extract['children'][] = $this->makeBookTreeExtract($child);
      }
    }
    return $extract;
  }

}
