<?php

namespace Drupal\skilling\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\skilling\Access\FilterUserInputInterface;
use Drupal\skilling\Access\SkillingAjaxSecurityInterface;
use Drupal\skilling\History;
use Drupal\skilling\SkillingConstants;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\skilling\SkillingCurrentUser;
use Symfony\Component\HttpFoundation\File\Exception\AccessDeniedException;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;

/**
 * Class SuggestionController.
 */
class SuggestionController extends ControllerBase {

  /**
   * Entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The Skilling current user service.
   *
   * @var \Drupal\skilling\SkillingCurrentUser
   */
  protected $skillingCurrentUser;

  /**
   * The Skilling history event recorder.
   *
   * @var \Drupal\skilling\History
   */
  protected $historyService;

  /**
   * Service to check AJAX calls.
   *
   * @var \Drupal\skilling\Access\SkillingAjaxSecurityInterface
   */
  protected $ajaxSecurityService;

  /**
   * Service to filter user input.
   *
   * @var \Drupal\skilling\Access\FilterUserInputInterface
   */
  protected $filterInputService;

  /**
   * Constructs a new controller object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   Entity type manager service.
   * @param \Drupal\skilling\SkillingCurrentUser $skillingCurrentUser
   *   Skilling current user service.
   * @param \Drupal\skilling\History $historyService
   *   Skilling history service.
   * @param \Drupal\skilling\Access\SkillingAjaxSecurityInterface $ajaxSecurityService
   *   Ajax security checking service.
   * @param \Drupal\skilling\Access\FilterUserInputInterface $filterInputService
   *   Input filter service.
   */
  public function __construct(
    EntityTypeManagerInterface $entityTypeManager,
    SkillingCurrentUser $skillingCurrentUser,
    History $historyService,
    SkillingAjaxSecurityInterface $ajaxSecurityService,
    FilterUserInputInterface $filterInputService
  ) {
    $this->entityTypeManager = $entityTypeManager;
    $this->skillingCurrentUser = $skillingCurrentUser;
    $this->historyService = $historyService;
    $this->ajaxSecurityService = $ajaxSecurityService;
    $this->filterInputService = $filterInputService;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    /* @noinspection PhpParamsInspection */
    return new static(
      $container->get('entity_type.manager'),
      $container->get('skilling.skilling_current_user'),
      $container->get('skilling.history'),
      $container->get('skilling.ajax_security'),
      $container->get('skilling.filter_user_input')
    );
  }

  /**
   * Save suggestion.
   *
   * Security note: get identifier and suggestion from client. Strip them.
   * Also, treat anything unexpected (e.g., node not found) as
   * a security exception.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   HTTP request.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The JSON response to the client.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\skilling\Exception\SkillingNotFoundException
   * @throws \Drupal\skilling\Exception\SkillingUnknownValueException
   */
  public function saveSuggestion(Request $request) {
    $passedBasicSecurity = $this->ajaxSecurityService->securityCheckAjaxRequest(
      $request,
      ['POST'],
      ['/skilling/save-suggestion']
    );
    if (!$passedBasicSecurity) {
      throw new AccessDeniedException('Access denied, sec fail in saveSuggestion');
    }
    if ($this->skillingCurrentUser->isAnonymous()) {
      throw new AccessDeniedException('Access denied, anon user in saveSuggestion');
    }
    // Get the user who owns the data.
    $suggestionNodeOwnerId = $this->skillingCurrentUser->id();
    // What the user typed.
    $suggestion = $request->get('suggestion');
    // Filter it.
    $suggestion = $this->filterInputService->filterUserContent($suggestion);
    /** @var \Drupal\node\NodeInterface $suggestionNode */
    $suggestionNode = $this->entityTypeManager
      ->getStorage('node')
      ->create(['type' => SkillingConstants::SUGGESTION_CONTENT_TYPE]);
    $title = $this->t(
      'Suggestion from user @uid',
      [
        '@uid' => $suggestionNodeOwnerId,
      ]
    );
    $suggestionNode->setTitle($title);
    $suggestionNode->setOwnerId($suggestionNodeOwnerId);
    // Set the suggestion text.
    $suggestionNode->get(SkillingConstants::FIELD_SUGGESTION)->value = $suggestion;
    $suggestionNode->save();
    // Record the event.
//    $this->historyService->createHistoryNode(
//      t('Create suggestion'),
//      SkillingConstants::HISTORY_EVENT_CREATE_SUGGESTION,
//      $this->skillingCurrentUser->id(), [], ''
//    );
    // Return JSON response.
    $result = [
      'status' => 'OK',
    ];
    return new JsonResponse($result);
  }

}
