<?php

namespace Drupal\skilling;

use Drupal\Component\Utility\Html;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\node\NodeInterface;
use Drupal\skilling\Exception\SkillingNotFoundException;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\skilling\Exception\SkillingValueMissingException;
use Drupal\user\Entity\User;

/**
 * Class History.
 */
class History {
  use StringTranslationTrait;

  /**
   * Has search already been recorded?
   *
   * @var bool
   */
  protected $searchRecorded = FALSE;

  /**
   * The Skilling current user service.
   *
   * @var \Drupal\skilling\SkillingCurrentUser
   */
  protected $skillingCurrentUser;

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * History item plugin factory.
   *
   * @var \Drupal\skilling\HistoryItemFactory
   */
  protected $historyItemFactory;

  /**
   * Constructs a new History object.
   *
   * @param \Drupal\skilling\SkillingCurrentUser $skillingCurrentUser
   *   The Skilling current user service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager service.
   * @param \Drupal\skilling\HistoryItemFactory $historyItemFactory
   *   History item plugin factory.
   */
  public function __construct(SkillingCurrentUser $skillingCurrentUser, EntityTypeManagerInterface $entityTypeManager, HistoryItemFactory $historyItemFactory) {
    $this->skillingCurrentUser = $skillingCurrentUser;
    $this->entityTypeManager = $entityTypeManager;
    $this->historyItemFactory = $historyItemFactory;
  }

  /**
   * Record the viewing of an event with the given node id. For students only.
   *
   * @param int $nid
   *   The node id.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\skilling\Exception\SkillingNotFoundException
   * @throws \Drupal\skilling\Exception\SkillingUnknownValueException
   * @throws \Drupal\skilling\Exception\SkillingValueMissingException
   */
  public function recordNidViewEvent($nid) {

    return;

    if (!$this->skillingCurrentUser->isStudent()) {
      return;
    }
    /** @var \Drupal\node\NodeInterface $node */
    $node = $this->entityTypeManager->getStorage('node')->load($nid);
    if ($node) {
      $this->recordNodeViewEvent($node);
    }
    else {
      throw new SkillingNotFoundException(
        Html::escape('Record view event: node not found: ' . $nid),
        __FILE__, __LINE__
      );
    }
  }

  /**
   * Record a node view event. For students only.
   *
   * @param \Drupal\node\NodeInterface $node
   *   The node viewed.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\skilling\Exception\SkillingUnknownValueException
   * @throws \Drupal\skilling\Exception\SkillingValueMissingException
   */
  public function recordNodeViewEvent(NodeInterface $node) {

    return;

    if (!$this->skillingCurrentUser->isStudent()) {
      return;
    }
    if (!$node) {
      throw new SkillingValueMissingException(
        Html::escape('Record view event: node missing'),
        __FILE__, __LINE__
      );
    }
    $eventCategoryName = $this->getHistoryViewEventCategoryForNode($node);
    if ($eventCategoryName === '') {
      return;
    }
    // Generate the title.
    $title = $this->t(
      'Viewed @title',
      ['@title' => $node->getTitle()]
    );
    $this->createHistoryNode($title, $eventCategoryName, $node->id(), []);
  }

  /**
   * Record a FiB answer.
   *
   * @param int $fibNid
   *   The FiB node id.
   * @param string $internalName
   *   The item's internal name.
   * @param string $response
   *   The student's answer.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\skilling\Exception\SkillingUnknownValueException
   */
  public function recordFibAnswer($fibNid, $internalName, $response) {

    return;

    $this->createHistoryNode(
      $this->t('Answered fill-in-the-blank'),
      SkillingConstants::HISTORY_EVENT_COMPLETE_FIB,
      $fibNid,
      [
        'fibNid' => $fibNid,
        'internalName' => Html::escape($internalName),
        'response' => Html::escape($response),
      ]
    );
  }

  /**
   * Record an MCQ answer.
   *
   * @param int $mcqNid
   *   Nid of the MCQ node.
   * @param string $internalName
   *   Internal name of the MCQ.
   * @param string $response
   *   The response the user chose.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\skilling\Exception\SkillingUnknownValueException
   */
  public function recordMcqAnswer($mcqNid, $internalName, $response) {

    return;

    $this->createHistoryNode(
      $this->t('Answered MCQ'),
      SkillingConstants::HISTORY_EVENT_COMPLETE_MCQ,
      $mcqNid,
      [
        'mcqNid' => $mcqNid,
        'internalName' => $internalName,
        'response' => $response,
      ],
      ''
    );
  }

  /**
   * Create a new history node.
   *
   * @param string $title
   *   Node title.
   * @param string $eventCategoryName
   *   The event category.
   * @param int $subjectId
   *   The id of the subject, e.g., node id.
   * @param array $details
   *   History event deets.
   * @param string $notes
   *   Instructor notes about the event.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\skilling\Exception\SkillingUnknownValueException
   */
  public function createHistoryNode($title, $eventCategoryName, $subjectId, array $details, $notes = '') {

    return;

    // Don't store search event more than once for this request.
    if ($eventCategoryName === SkillingConstants::HISTORY_EVENT_SEARCH) {
      if ($this->searchRecorded) {
        return;
      }
      $this->searchRecorded = TRUE;
    }
    $historyItem = $this->historyItemFactory->makeNodelessHistoryEventObject($eventCategoryName);
    // Login events need special handling. Hook code sends id of logging in user
    // in $subjectId.
    if ($eventCategoryName === SkillingConstants::HISTORY_EVENT_LOG_IN) {
      $historyItem->loggingInUserId = $subjectId;
    }
    // Should record this event?
    if (!$historyItem->isLogged()) {
      return;
    }
    // User this is for.
    // Want events for message to students to use the student id as the
    // event target. The call to the current method from tne
    // contact form sends the student's id in $subjectId.
    if ($eventCategoryName === SkillingConstants::HISTORY_EVENT_INSTRUCTOR_SENT_MESSAGE_TO_STUDENT) {
      $uid = $subjectId;
    }
    elseif ($eventCategoryName === SkillingConstants::HISTORY_EVENT_LOG_IN) {
      // Logins use the id of the user logging in, not the current user (anon).
      $uid = $subjectId;
    }
    else {
      $uid = $historyItem->getUidEventIsFor();
    }
    // Add history node.
    /** @var \Drupal\node\Entity\Node $node */
    $node = $this->entityTypeManager
      ->getStorage('node')
      ->create(['type' => SkillingConstants::HISTORY_CONTENT_TYPE]);
    $node->set('title', $title);
    // Owner of node is the user it is about.
    $node->set('uid', $uid);
    // Event type.
    $node->set('field_event_type_name', $eventCategoryName);
    /* @noinspection PhpUndefinedFieldInspection */
    $node->field_user->target_id = $uid;
    // field_when set with default value of current time.
    if (!is_null($subjectId)) {
      $node->field_subject_node->target_id = $subjectId;
    }
    // Serialize event detail array for storage.
    if (count($details) > 0) {
      $node->set('field_details', serialize($details));
    }
    if ($notes) {
      $node->set('field_notes', $notes);
    }
    $node->field_notes->format = SkillingConstants::SKILLING_TEXT_FORMAT;
    $node->save();
  }

  /**
   * Find the last lesson viewed by an account.
   *
   * @param \Drupal\user\Entity\User $account
   *   Check this user.
   *
   * @return int|null
   *   Lesson id or null if none.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function findLastLessonViewedByAccount(User $account) {

    return NULL;

    // Only for students.
    if (!$account->hasRole(SkillingConstants::CLASS_ROLE_STUDENT)) {
      return NULL;
    }
    $uid = $account->id();
    $historyNodeIds = $this->entityTypeManager
      ->getStorage('node')
      ->getQuery()
      ->condition('type', SkillingConstants::HISTORY_CONTENT_TYPE)
      // Is published?
      ->condition('status', TRUE)
      // Of view lesson.
      ->condition(
        'field_event_type_name',
        SkillingConstants::HISTORY_EVENT_VIEW_LESSON
         )
      // Sort by date.
      ->sort('created', 'DESC')
      // Get the first record.
      ->range(0, 1)
      ->execute();
    // Found anything?
    if (count($historyNodeIds) === 0) {
      // Nothing.
      return NULL;
    }
    $historyNodeId = reset($historyNodeIds);
    // Load the node that was referred to.
    $historyNode = $this->entityTypeManager->getStorage('node')
      ->load($historyNodeId);
    if (!$historyNode) {
      return NULL;
    }
    $lessonNid = $historyNode->get('field_subject_node')->target_id;
    // Return the id of the lesson node.
    return $lessonNid;
  }

  /**
   * Get the history event category for viewing a node.
   *
   * @param \Drupal\node\NodeInterface $node
   *   Node being viewed.
   *
   * @return string|null
   *   History event category, NULL if not to be recorded.
   */
  public function getHistoryViewEventCategoryForNode(NodeInterface $node) {

    return NULL;

    switch ($node->bundle()) {
      case SkillingConstants::LESSON_CONTENT_TYPE:
        $eventCategory = SkillingConstants::HISTORY_EVENT_VIEW_LESSON;
        break;

      case SkillingConstants::DESIGN_PAGE_CONTENT_TYPE:
        $eventCategory = SkillingConstants::HISTORY_EVENT_VIEW_DESIGN_PAGE;
        break;

      case SkillingConstants::EXERCISE_CONTENT_TYPE:
        $eventCategory = SkillingConstants::HISTORY_EVENT_VIEW_EXERCISE;
        break;

      case SkillingConstants::EXERCISE_SUBMISSION_CONTENT_TYPE:
        $eventCategory = SkillingConstants::HISTORY_EVENT_VIEW_EXERCISE_SUBMISSION;
        break;

      case SkillingConstants::CLASS_CONTENT_TYPE:
        $eventCategory = SkillingConstants::HISTORY_EVENT_VIEW_YOUR_CLASS;
        break;

      case SkillingConstants::EVENT_CONTENT_TYPE:
        $eventCategory = SkillingConstants::HISTORY_EVENT_VIEW_EVENT;
        break;

      case SkillingConstants::PATTERN_CONTENT_TYPE:
        $eventCategory = SkillingConstants::HISTORY_EVENT_VIEW_PATTERN;
        break;

      case SkillingConstants::PRINCIPLE_CONTENT_TYPE:
        $eventCategory = SkillingConstants::HISTORY_EVENT_VIEW_PRINCIPLE;
        break;

      case SkillingConstants::MODEL_CONTENT_TYPE:
        $eventCategory = SkillingConstants::HISTORY_EVENT_VIEW_PRINCIPLE;
        break;

      default:
        $eventCategory = NULL;
        break;
    }
    return $eventCategory;
  }

}
