<?php

namespace Drupal\skilling\Plugin;

use Drupal\Component\Plugin\PluginBase;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\node\NodeInterface;
use Drupal\skilling\Access\FilterUserInput;
use Drupal\skilling\Access\FilterUserInputInterface;
use Drupal\skilling\Access\SkillingAjaxSecurityInterface;
use Drupal\skilling\SkillingCurrentUser;

/**
 * Base class for Skilling history item plugins.
 *
 * The plugins wrap history nodes, giving methods that spit out
 * renderables and XML for the deets field.
 */
class HistoryItemBase extends PluginBase implements HistoryItemInterface {

  use StringTranslationTrait;

  /**
   * The history node.
   *
   * @var \Drupal\node\NodeInterface
   */
  protected $historyEventNode;

  /**
   * Current user service.
   *
   * Factories that make HistoryItems have to set this.
   *
   * @var \Drupal\skilling\SkillingCurrentUser
   */
  protected $skillingCurrentUser = FALSE;

  /**
   * Uid of user who is logging in.
   *
   * This is a special case for login events only, when the current user
   * is anon, but we want to record a history event for students. The code is
   * a bit hacky, but is OK for the one special case.
   *
   * @var int
   */
  public $loggingInUserId;

  /**
   * Service to check AJAX calls.
   *
   * @var \Drupal\skilling\Access\SkillingAjaxSecurityInterface
   */
  protected $ajaxSecurityService;

  /**
   * Service to filter user input.
   *
   * @var \Drupal\skilling\Access\FilterUserInputInterface
   */
  protected $filterInputService;

  /**
   * Get the @description property from the plugin annotation, and return it.
   *
   * @return string
   *   Description property value.
   */
  public function description() {
    return $this->pluginDefinition['description'];
  }

  /**
   * Get the @event_type_name property from the plugin annotation.
   *
   * @return string
   *   The name, e.g., logout.
   */
  public function getEventTypeName() {
    return $this->pluginDefinition['event_type_name'];
  }

  /**
   * Format a history error message for reporting.
   *
   * @param string $message
   *   Error message to report.
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup|mixed
   *   The report.
   */
  public function formatHistoryItemError($message) {
    $result = $this->t('History event error: @m', ['@t' => $message]);
    return $result;
  }

  /**
   * Get a renderable array for showing deets of a history event.
   *
   * Default is to return MT array. No deets to show.
   *
   * @return array
   *   The renderable.
   */
  public function getDetailsRenderable() {
    return [];
  }

  /**
   * Get XML representation of the deets of a history event.
   *
   * Default is to return MT string. No deets to represent.
   *
   * @return string
   *   The XML.
   */
  public function getDetailsXml() {
    return '';
  }

  /**
   * Set the node being wrapped.
   *
   * @param \Drupal\node\NodeInterface $historyEventNode
   *   The node.
   */
  public function setHistoryEventNode(NodeInterface $historyEventNode) {
    $this->historyEventNode = $historyEventNode;
  }

  /**
   * Get the node being wrapped.
   *
   * @return \Drupal\node\NodeInterface
   *   The node.
   */
  public function getHistoryEventNode() {
    return $this->historyEventNode;
  }

  /**
   * Set the current user.
   *
   * @param \Drupal\skilling\SkillingCurrentUser $skillingCurrentUser
   *   The logged in user.
   */
  public function setCurrentUser(SkillingCurrentUser $skillingCurrentUser) {
    $this->skillingCurrentUser = $skillingCurrentUser;
  }

  /**
   * Set the AJAX security service.
   *
   * @param \Drupal\skilling\Access\SkillingAjaxSecurityInterface $ajaxSecurityService
   *   The service.
   */
  public function setAjaxSecurityService(SkillingAjaxSecurityInterface $ajaxSecurityService) {
    $this->ajaxSecurityService = $ajaxSecurityService;
  }

  /**
   * Set the user input filter service.
   *
   * @param \Drupal\skilling\Access\FilterUserInputInterface $filterInputService
   *   The service.
   */
  public function setFilterUserInputService(FilterUserInputInterface $filterInputService) {
    $this->filterInputService = $filterInputService;
  }

  /**
   * Is the event to be logged in history?
   *
   * Default method logs for students only.
   *
   * @return bool
   *   Whether to log the event.
   */
  public function isLogged() {
    if (!$this->skillingCurrentUser) {
      $this->skillingCurrentUser = \Drupal::service('skilling.skilling_current_user');
    }
    return $this->skillingCurrentUser->isStudent();
  }

  /**
   * Get the uid the event is for.
   *
   * Is almost always the uid of the current user. Plugin implementations
   * can do something different if they want.
   *
   * @return int
   *   The uid.
   */
  public function getUidEventIsFor() {
    if (!$this->skillingCurrentUser) {
      $this->skillingCurrentUser = \Drupal::service('skilling.skilling_current_user');
    }
    return $this->skillingCurrentUser->id();
  }

  /**
   * Prepare an array to be saved as XML.
   *
   * @param array $itemsIn
   *   The data to prepare.
   *
   * @return array
   *   Prepared data.
   */
  public function prepareForXml(array $itemsIn) {
    $itemsOut = [];
    foreach ($itemsIn as $itemIn) {
      if (is_int($itemIn)) {
        $itemsOut[] = (int) $itemIn;
      }
      elseif (is_float($itemIn)) {
        $itemsOut[] = (float) $itemIn;
      }
      elseif (is_bool($itemIn)) {
        $itemsOut[] = $itemsIn ? TRUE : FALSE;
      }
      else {
        $itemsOut[] = htmlspecialchars($itemIn, ENT_XML1 | ENT_QUOTES, 'UTF-8');
      }
    }
    return $itemsOut;
  }
}
