<?php
namespace Drupal\skilling\Plugin\SkillingCustomTag;

use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\file\Entity\File;
use Drupal\node\Entity\Node;
use Drupal\skilling\Exception\SkillingException;
use Drupal\skilling\Plugin\SkillingCustomTagBase;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\skilling\SkillingConstants;
use Drupal\skilling\SkillingCurrentUser;
use Drupal\skilling\Submissions;
use Drupal\skilling\Utilities as SkillingUtilities;
use Drupal\skilling\NidBag;

/**
 * Provides an exercise tag.
 *
 * @SkillingCustomTag(
 *  id = "skilling_exercise_custom_tag",
 *  tag = "exercise",
 *  admin_label = @Translation("Exercise"),
 *  description = @Translation("Inserts an exercise into a lesson."),
 *  has_close_tag = false
 * )
 */
class ExerciseTag extends SkillingCustomTagBase {

  use StringTranslationTrait;

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The Skilling utilities service.
   *
   * @var \Drupal\skilling\Utilities
   */
  protected $skillingUtilities;

  /**
   * The Skilling current user service.
   *
   * Provides data about the current user, e.g., roles.
   *
   * @var \Drupal\skilling\SkillingCurrentUser
   */
  protected $currentUser;

  /**
   * A service for tracking node ids.
   *
   * Helps track nodes inserted in lessons by the parser. Used to
   * update entity reference field on inserted nodes.
   *
   * @var \Drupal\skilling\NidBag
   */
  protected $nidBag;

  /**
   * Submissions service.
   *
   * @var \Drupal\skilling\Submissions
   */
  protected $submissionsService;

  /**
   * ExerciseTag constructor.
   *
   * @param array $configuration
   *   Plugin configuration.
   * @param string $pluginId
   *   Plugin id.
   * @param $pluginDefinition
   *   Plugin definition.
   */
  public function __construct(array $configuration, $pluginId, $pluginDefinition) {
    parent::__construct($configuration, $pluginId, $pluginDefinition);
    $this->entityTypeManager = \Drupal::service('entity_type.manager');
    $this->skillingUtilities = \Drupal::service('skilling.utilities');
    $this->nidBag = \Drupal::service('skilling.nid_bag');
    $this->currentUser = \Drupal::service('skilling.skilling_current_user');
    $this->submissionsService = \Drupal::service('skilling.submissions');
  }

  /**
   * Process content, with options.
   *
   * @param string $content
   *   Content to process.
   * @param array $options
   *   Options.
   *
   * @param \Drupal\node\Entity\Node|null $nodeWithTag
   *
   * @return string
   *   Rendered result.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\skilling\Exception\SkillingException
   * @throws \Drupal\skilling\Exception\SkillingParserException
   */
  public function processTag($content = '', array $options = NULL, Node $nodeWithTag = NULL) {
    // Find the name of the exercise.
    // Allow for some mistyping of option name by authors.
    $internalName = $this->findOption(
      $options, SkillingConstants::INTERNAL_NAME_ALTS
    );
    if ($internalName === '') {
      return $this->formatCustomTagError(
        $this->t('Exercise tag needs internal_name option.')
      );
    }
    // Find the exercise.
    // This code is slightly different from
    // Utilities::getPublishedExerciseWithInternalName. It does not
    // throw an exception when an exericse is unpublished.
    $exerciseIds = $this->entityTypeManager->getStorage('node')
      ->getQuery()
      // Is an exercise.
      ->condition('type', SkillingConstants::EXERCISE_CONTENT_TYPE)
      // Has the right internal name.
      ->condition('field_internal_name', $internalName)
      ->execute();
    if (!$exerciseIds) {
      return $this->formatCustomTagError(
        $this->t('Exercise with internal_name @n not found.', ['@n' => $internalName])
      );
    }
    if (count($exerciseIds) !== 1) {
      return $this->formatCustomTagError(
        $this->t(
          'Unexpected number of exercises with internal_name @n found.',
          ['@n' => $internalName])
      );
    }
    // Load the exercise.
    $exerciseId = reset($exerciseIds);
    /** @var \Drupal\node\Entity\Node $exercise */
    $exercise = $this->entityTypeManager->getStorage('node')
      ->load($exerciseId);
    // Should not happen.
    if (!$exercise) {
      return $this->formatCustomTagError(
        $this->t('Exercise with id @id not found.', ['@id' => $exerciseId])
      );
    }
    // If not published, return nothing.
    if (!$exercise->isPublished()) {
      return '';
    }
    // Drop the nid into a bag, in case others want it.
    $this->nidBag->addToBag(SkillingConstants::NID_POCKET_EXERCISES, $exerciseId);
    // Get body as shown in exercise's teaser.
    $summary = $exercise->body->summary;
    // Get the full body.
    $task = $exercise->body->value;
    $task = $this->parser->parse($task);
    $renderableSubmissionLinks = $this->submissionsService->createSubmissionLinks($exercise);
    /* @noinspection PhpUndefinedFieldInspection */
    $renderable = [
      '#theme' => 'insert_exercise',
      '#title' => $exercise->getTitle(),
      '#summary' => $summary,
      '#task' => $task,
      // Exercise nid is passed to submission popups, that pass it to
      // code in exercise-links JS library that closes popups,
      // and adjusts submission links on the current page.
      '#exercise_nid' => $exercise->id(),
      '#internal_name' => $exercise->field_internal_name->value,
      '#submission_links' => $renderableSubmissionLinks,
      '#attached' => [
        'library' => 'skilling/exercise-links',
        'drupalSettings' => [
          'skillingFloatQueryParam' => SkillingConstants::FLOATER_QUERY_STRING,
        ],
      ],
//      '#attached' => [
//        'library' => 'skilling/exercise-links',
//        'drupalSettings' => [
//          'skillingFloatQueryParam' => SkillingConstants::FLOATER_QUERY_STRING,
//        ],
//      ],
    ];
    $rendered = \Drupal::service('renderer')->render($renderable);
    return $rendered;
  }

}
