<?php

namespace Drupal\skilling;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\node\NodeInterface;
use Drupal\skilling\Exception\SkillingWrongTypeException;
use Drupal\skilling\Utilities;
use Drupal\skilling\CurrentClass;
use Drupal\skilling\SkillingCurrentUser;
use Drupal\Core\Messenger\MessengerInterface;

/**
 * Class SkillingClass.
 */
class SkillingClass {

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Utilities service.
   *
   * @var \Drupal\skilling\Utilities
   */
  protected $skillingUtilities;

  /**
   * The current class service.
   *
   * @var \Drupal\skilling\CurrentClass
   */
  protected $skillingCurrentClass;

  /**
   * The Skilling current user service.
   *
   * @var \Drupal\skilling\SkillingCurrentUser
   */
  protected $skillingSkillingCurrentUser;

  /**
   * The messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Constructs a new SkillingClass object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   Entity type manager service.
   * @param \Drupal\skilling\Utilities $skillingUtilities
   *   Utilities service.
   * @param \Drupal\skilling\SkillingCurrentUser $skillingCurrentUser
   *   The Skilling current user service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   */
  public function __construct(
    EntityTypeManagerInterface $entityTypeManager,
    Utilities $skillingUtilities,
    SkillingCurrentUser $skillingCurrentUser,
    MessengerInterface $messenger
  ) {
    $this->entityTypeManager = $entityTypeManager;
    $this->skillingUtilities = $skillingUtilities;
    $this->skillingSkillingCurrentUser = $skillingCurrentUser;
    $this->messenger = $messenger;
  }

  /**
   * Get the instructors for a class.
   *
   * @param \Drupal\node\NodeInterface $class
   *   The class.
   *
   * @return \Drupal\user\Entity\User[]
   *   User objects.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\skilling\Exception\SkillingWrongTypeException
   */
  public function getInstructorsForClass(NodeInterface $class) {
    if ($class->bundle() !== SkillingConstants::CLASS_CONTENT_TYPE) {
      throw new SkillingWrongTypeException(
        'Expected class node, got ' . $class->bundle(),
        __FILE__, __LINE__
      );
    }
    // Find enrollments for the class, with instructor role.
    $enrollmentIds = $this->entityTypeManager
      ->getStorage('node')
      ->getQuery()
      ->condition('type', SkillingConstants::ENROLLMENT_CONTENT_TYPE)
      // Enrollment is published.
      ->condition('status', TRUE)
      // For the class.
      ->condition('field_class', $class->id())
      // Class is published.
      ->condition('field_class.entity.status', 1)
      // Is an instructor.
      ->condition('field_class_roles', [SkillingConstants::CLASS_ROLE_INSTRUCTOR], 'IN')
      ->execute();
    $enrollments = $this->entityTypeManager
      ->getStorage('node')
      ->loadMultiple($enrollmentIds);
    $instructorIds = [];
    /** @var \Drupal\node\Entity\Node $enrollment */
    foreach ($enrollments as $enrollment) {
      $instructorIds[] = $enrollment->get('field_user')->target_id;
    }
    // Load instructor's user objects.
    /** @var \Drupal\user\Entity\User[] $instructors */
    $instructors = $this->entityTypeManager
      ->getStorage('user')
      ->loadMultiple($instructorIds);
    return $instructors;
  }



}
