<?php

namespace Drupal\skilling;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use DateInterval;
use DateTime;

/**
 * Class Timeline.
 */
class Timeline {

  use StringTranslationTrait;

  /**
   * Entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Current class service.
   *
   * @var \Drupal\skilling\CurrentClass
   */
  protected $currentClass;

  /**
   * Skilling current user service.
   *
   * @var \Drupal\skilling\SkillingCurrentUser
   */
  protected $currentUser;

  /**
   * Submissions service.
   *
   * @var \Drupal\skilling\Submissions
   */
  protected $submissionsService;

  /**
   * Constructs a new Timeline object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   Entity type manager service.
   * @param \Drupal\skilling\CurrentClass $skilling_current_class
   *   Current class service.
   * @param \Drupal\skilling\SkillingCurrentUser $skilling_skilling_current_user
   *   Skilling current user service.
   * @param \Drupal\skilling\Submissions $submissionsService
   *   Submissions service.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, CurrentClass $skilling_current_class, SkillingCurrentUser $skilling_skilling_current_user, Submissions $submissionsService) {
    $this->entityTypeManager = $entity_type_manager;
    $this->currentClass = $skilling_current_class;
    $this->currentUser = $skilling_skilling_current_user;
    $this->submissionsService = $submissionsService;
  }

  /**
   * Make a timeline for the current user.
   *
   * Include all events from the
   * user's current class. For exercises, submissions for exercises that
   * are not on the calendar do not show.
   *
   * @return array
   *   Form: [problemMessage,eventsForTemplate]
   *   Ready to be wrapped and passed to the Twig template.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Exception
   */
  public function makeTimeline() {
    $problemMessage = '';
    $eventsForTemplate = [];
    if ($this->currentUser->isAnonymous()) {
      $problemMessage = $this->t('Calendar only available when logged in.');
      return [$problemMessage, $eventsForTemplate];
    }
    // Get the class.
    /** @var \Drupal\node\Entity\Node $class */
    $class = $this->currentClass->getCurrentClass();
    if (!$class) {
      $problemMessage = $this->t('No current class.');
      return [$problemMessage, $eventsForTemplate];
    }
    // There is a class.
    // Get the calendar for the class.
    /* @var \Drupal\node\Entity\Node $calendar */
    $calendar = NULL;
    /* @noinspection PhpUndefinedFieldInspection */
    $calendarId = $class->field_calendar->target_id;
    if (!$calendarId) {
      $problemMessage = $this->t('No calendar for class.');
      return [$problemMessage, $eventsForTemplate];
    }
    $calendar = $this->entityTypeManager
      ->getStorage('node')
      ->load($calendarId);
    if (!$calendar || !$calendar->isPublished()) {
      $problemMessage = $this->t('Calendar not available.');
      return [$problemMessage, $eventsForTemplate];
    }
    // Test that should never pass.
    /* @noinspection PhpUndefinedFieldInspection */
    if (!isset($calendar->field_days->value) || !$calendar->field_days->value) {
      $problemMessage = $this->t('Calendar end day not found.');
      return [$problemMessage, $eventsForTemplate];
    }
    // Get the calendar events, and their events.
    $calendarEvents = [];
    $events = [];
    foreach ($calendar->field_calendar_events as $calendarEventFieldValue) {
      $calendarEventParagraph = $calendarEventFieldValue->entity;
      $calendarEvents[] = $calendarEventParagraph;
      /** @var \Drupal\node\NodeInterface $event */
      $event = $calendarEventParagraph->field_event->entity;
      if ($event) {
        $events[$event->id()] = $event;
      }
    }
    if (count($events) === 0) {
      $problemMessage = $this->t('No events found.');
      return [$problemMessage, $eventsForTemplate];
    }
    // Run through the calendar event nodes. Find the associated event.
    // Make a new array combining the event data with the day
    // it happens.
    // Also extract the exercise ids, for loading later.
    $eventList = [];
    $exerciseNids = [];
    foreach ($calendarEvents as $calendarEvent) {
      /* @noinspection PhpUndefinedFieldInspection */
      $eventRef = $calendarEvent->field_event;
      if (isset($eventRef->target_id) && $eventRef->target_id) {
        $eventNid = $eventRef->target_id;
        /* @var \Drupal\node\Entity\Node $event */
        $event = $events[$eventNid];
        if ($event->isPublished()) {
          /* @noinspection PhpUndefinedFieldInspection */
          $day = $calendarEvent->field_day->value;
          // Check just in case.
          if ($day) {
            $eventTitle = $event->getTitle();
            /* @noinspection PhpUndefinedFieldInspection */
            $eventSummary = isset($event->body->summary) && $event->body->summary
              ? $event->body->summary : '';
            /* @noinspection PhpUndefinedFieldInspection */
            $eventType = $event->field_event_type->value;
            /* @noinspection PhpUndefinedFieldInspection */
            $exerciseNid = ($eventType == SkillingConstants::EVENT_TYPE_EXERCISE_DUE)
              ? $event->field_exercise->target_id : 0;
            $eventList[] = [
              'event_nid' => $eventNid,
              'event_type' => $eventType,
              'day' => (int) $day,
              'title' => $eventTitle,
              'description' => $eventSummary,
              'exercise_nid' => $exerciseNid,
            ];
            // If this is an exercise event, remember the exercise's nid.
            if ($exerciseNid !== 0) {
              $exerciseNids[] = $exerciseNid;
            }
          } //End day is set.
        } //End event is published.
      } //End event ref is set.
    } //End for each calendar event.
    // Any of the events exercises?
    if (count($exerciseNids) > 0) {
      // Yes.
      // Load the exercises referenced by the events.
      // NB! ONLY exercises listed in events for the calendar.
      // Not all exercises, or exercises from submissions for the user.
      // Includes exercises with no submissions from the user.
      $exercisesRaw = $this->entityTypeManager->getStorage('node')
        ->loadMultiple($exerciseNids);
      // Remove unpublished exercises.
      $exercises = [];
      $exerciseNids = [];
      /* @var \Drupal\node\Entity\Node $exercise */
      foreach ($exercisesRaw as $exercise) {
        if ($exercise->isPublished()) {
          $exercises[$exercise->id()] = $exercise;
          $exerciseNids[] = $exercise->id();
        }
      }
      // Add the exercise titles to the event list data.
      $eventCount = count($eventList);
      for ($index = 0; $index < $eventCount; $index++) {
        $eventItem = $eventList[$index];
        $exerciseNid = $eventItem['exercise_nid'];
        $exerciseTitle = '';
        $exerciseInternalName = '';
        // Nid is 0 for non-exercise events.
        if ($exerciseNid !== 0 && array_key_exists($exerciseNid, $exercises)) {
          /* @var \Drupal\node\Entity\Node $exercise */
          $exercise = $exercises[$exerciseNid];
          $exerciseTitle = $exercise->getTitle();
          /* @noinspection PhpUndefinedFieldInspection */
          $exerciseInternalName = $exercise->field_internal_name->value;
        }
        $eventList[$index]['exercise_title'] = $exerciseTitle;
        $eventList[$index]['exercise_internal_name'] = $exerciseInternalName;
        // Add the submission status - MT to start with.
        $eventList[$index]['exercise_submission_status'] = '';
      } // End loop over eventList elements.
      // Check submissions for students only.
      if (!$this->currentUser->isStudent()) {
        // For non-students, set special submission status.
        for ($index = 0; $index < $eventCount; $index++) {
          $eventItem = $eventList[$index];
          $exerciseNid = $eventItem['exercise_nid'];
          // Nid is 0 for non-exercise events.
          if ($exerciseNid !== 0) {
//            if (isset($exercisesSubmissionSummary[$exerciseNid])) {
              $eventList[$index]['exercise_submission_status']
                = SkillingConstants::EXERCISE_SUBMISSION_NA;
//            }
          }
        } //End for each event.
      }
      else {
        // Load the submissions for the exercises.
        $submissionIds = $this->entityTypeManager->getStorage('node')
          ->getQuery()
          ->condition('type', 'submission')
          // For the current user.
          ->condition('field_user', $this->currentUser->id())
          // Published submissions.
          ->condition('status', 1)
          // One of the exercises in question.
          ->condition('field_exercise.target_id', $exerciseNids, 'IN')
          // Exercise is published. Extra check.
          ->condition('field_exercise.entity.status', 1)
          ->execute();
        $submissions = $this->entityTypeManager->getStorage('node')
          ->loadMultiple($submissionIds);
        // Find out the submission status of each exercise.
        // Returns array indexed by exercise nid.
        $exercisesSubmissionSummary = $this->submissionsService
          ->getSubmissionStatusForStudentExercisesSubmissions(
            $this->currentUser, $exerciseNids, $submissions
          );
        // Add submission status to event data.
        for ($index = 0; $index < $eventCount; $index++) {
          $eventItem = $eventList[$index];
          $exerciseNid = $eventItem['exercise_nid'];
          // Nid is 0 for non-exercise events.
          if ($exerciseNid !== 0) {
            if (isset($exercisesSubmissionSummary[$exerciseNid])) {
              $eventList[$index]['exercise_submission_status']
                = $exercisesSubmissionSummary[$exerciseNid];
            }
          }
        } //End for each event.
      } //End user is a student.
    } //End there are exercises.
    // Each event will have an HTML link. For exercises, the link
    // will be to the place in a lesson where the exercises is shown, or,
    // if the exercise is not inserted into a lesson, to the
    // exercise itself. For other events, the link will be to the event
    // deets.
    $eventCount = count($eventList);
    for ($index = 0; $index < $eventCount; $index++) {
      $eventItem = $eventList[$index];
      $eventType = $eventItem['event_type'];
      if ($eventType === SkillingConstants::EVENT_TYPE_SHOW_TEXT) {
        // For show text events, link points to the event.
        $linkedObjectType = 'event';
        $nidForLink = $eventItem['event_nid'];
      }
      else {
        // Event is an exercise.
        // For exercise due events, link points to the page in which the
        // exercise is embedded.
        // If the exercise is not embedded in a page, the event link
        // points to the exercise.
        $exerciseNid = $eventItem['exercise_nid'];
        if ($exerciseNid === 0 || !isset($exercises[$exerciseNid])) {
          // Should never happen.
          continue;
        }
        /* @var \Drupal\node\Entity\Node $exercise */
        /* @noinspection PhpUndefinedVariableInspection */
        $exercise = isset($exercises[$exerciseNid])
          ? $exercises[$exerciseNid] : NULL;
        if (!$exercise) {
          continue;
        }
        if (!$exercise->isPublished()) {
          // Already checked, but just in case.
          continue;
        }
        // Find lesson the exercise is inserted in, if there is one.
        /* @noinspection PhpUndefinedFieldInspection */
        $lessonReferences = $exercise->field_where_referenced->referencedEntities();
        if ($lessonReferences && count($lessonReferences) > 0) {
          // There is a lesson that has the exercise inserted.
          // Link to the lesson. The exercise's internal name is the
          // page anchor for the link.
          $linkedObjectType = 'lesson';
          $nidForLink = current($lessonReferences)->id();
        }
        else {
          // No lesson includes the exercise. Link to the exercise itself.
          $linkedObjectType = 'exercise';
          $nidForLink = $exerciseNid;
        } // End exercise not inserted into a lesson.
      }
      $eventList[$index]['linked_object_type'] = $linkedObjectType;
      $eventList[$index]['linked_nid'] = $nidForLink;
      if ($eventType === SkillingConstants::EVENT_TYPE_SHOW_TEXT) {
        $eventList[$index]['marker'] = '&nbsp;';
      }
    } // End loop across event items.
    // Work out the max day.
    // Could be the number of days in the calendar, or there could be an event
    // with a greater day.
    /* @noinspection PhpUndefinedFieldInspection */
    $lastDay = $calendar->field_days->value;
    foreach ($eventList as $eventItem) {
      $eventDay = $eventItem['day'];
      if ($eventDay > $lastDay) {
        $lastDay = $eventDay;
      }
    }
    // Sort by day.
    usort($eventList, function ($a, $b) {
      $aDay = $a['day'];
      $bDay = $b['day'];
      if ($aDay == $bDay) {
        return 0;
      }
      elseif ($aDay > $bDay) {
        return 1;
      }
      else {
        return -1;
      }
    });
    $now = new DateTime();
    // Date the class starts. May be in the middle of a week.
    $classStartDate = new DateTime($class->field_when_starts->value);
    // Day of the week the class starts, 0=Sunday.
    $classStartDateDayOfWeek = (int) $classStartDate->format('w');
    // Displayed calendar starts from the beginning of the week
    // that contains the class start date.
    $displayCalendarStartDate = clone $classStartDate;
    $displayCalendarStartDate->sub(
      new DateInterval('P' . $classStartDateDayOfWeek . 'D')
    );
    $weeks = [];
    // Loop across events.
    foreach ($eventList as $event) {
      $eventDay = $event['day'];
      // Days from the beginning of week 1 that the event happens.
      $daysFromDisplayCalendarStart = ($eventDay - 1) + $classStartDateDayOfWeek;
      // Week the event is in, 1 based.
      $eventWeek = (int) ($daysFromDisplayCalendarStart / 7 + 1);
      if (!isset($weeks[$eventWeek])) {
        $weeks[$eventWeek] = [];
      }
      // The event's day of the week. 0 is Sunday.
      $eventDayOfWeek = $daysFromDisplayCalendarStart % 7;
      if (!isset($weeks[$eventWeek][$eventDayOfWeek])) {
        // Compute the components of the event's date for display.
        $interval = new DateInterval('P' . ($eventDay - 1) . 'D');
        $eventDate = clone $classStartDate;
        $eventDate->add($interval);
        $dateDisplayStuff = [
          'month' => $eventDate->format('M'),
          'dow' => $eventDate->format('D'),
          'day' => $eventDate->format('j'),
          'today' => 'no',
          'is_past' => $eventDate < $now,
        ];
        $weeks[$eventWeek][$eventDayOfWeek] = [
          'events' => [],
          'date_display' => $dateDisplayStuff,
        ];
      }
      $weeks[$eventWeek][$eventDayOfWeek]['events'][] = $event;
    }
    // Add today to the $weeks array.
    // Difference between now and display calendar's start date.
    $diff = $displayCalendarStartDate->diff($now, TRUE);
    // Difference expressed in days.
    $daysDiff = (int) $diff->format('%d');
    // The week now is in, 1 based.
    $nowWeek = (int) ($daysDiff / 7 + 1);
    // Now's day of the week, 0=Sunday.
    $nowDayOfWeek = $daysDiff % 7;
    // Is there another event today?
    if (!isset($weeks[$nowWeek])) {
      $weeks[$nowWeek] = [];
    }
    if (!isset($weeks[$nowWeek][$nowDayOfWeek])) {
      // No. Create a new entry in $weeks.
      $dateDisplayStuff = [
        'month' => $now->format('M'),
        'dow' => $now->format('D'),
        'day' => $now->format('j'),
        'today' => 'yes',
      ];
      $weeks[$nowWeek][$nowDayOfWeek] = [
        'events' => [],
        'date_display' => $dateDisplayStuff,
      ];
    }
    else {
      // Other events for today. Mark $weeks entry as for today.
      $weeks[$nowWeek][$nowDayOfWeek]['date_display']['today'] = 'yes';
    }
    ksort($weeks);
    $sortedWeeks = [];
    foreach ($weeks as $key => $row) {
      ksort($row);
      $sortedWeeks[$key] = $row;
    }
    return [$problemMessage, $sortedWeeks];





    // Make the day list to pass to the template.
    $eventsForTemplate = [];
    for ($day = 1; $day <= $lastDay; $day++) {
      $eventsForTemplate[$day] = [
        'day' => $day,
        'events' => [],
      ];
    }
    // Add dates for the first day, and last day.
    /* @noinspection PhpUndefinedFieldInspection */
    $startDate = $class->field_when_starts->value;
    $startDate = new DateTime($startDate);
    $lastDate = clone $startDate;
    $lastDate->add(new DateInterval('P' . $lastDay . 'D'));
    $eventsForTemplate[1]['display_date'] = $this->formatTimelineDate($startDate);
    $eventsForTemplate[$lastDay]['display_date'] = $this->formatTimelineDate($lastDate);
    // Work out the display date for each day.
    $oneDayInterval = new DateInterval('P1D');
    $tempDate = clone $startDate;
    $tempDate->add($oneDayInterval);
    for ($day = 2; $day < $lastDay; $day++) {
      $displayDate = $this->formatTimelineDate($tempDate);
      $eventsForTemplate[$day]['display_date'] = $displayDate;
      $tempDate->add($oneDayInterval);
    }
    // What is today, expressed as number of days from start of course?
    $now = new DateTime();
    $todayAsDayNumber = $now->diff($startDate)->days + 1;
    // Add the events to the list for the template.
    foreach ($eventList as $eventItem) {
      $day = $eventItem['day'];
      // Add is-past flag to events.
      $eventItem['is_past'] = $day < $todayAsDayNumber;
      // Add event to list.
      $eventsForTemplate[$day]['events'][] = $eventItem;
    }
    // Simplify the display dates when there is no event, and it
    // isn't Sunday. Leave the first and last days alone.
    for ($day = 2; $day < $lastDay; $day++) {
      // Mark today.
      $eventsForTemplate[$day]['is_today'] = ($day === $todayAsDayNumber);
      $eventDisplayDate = $eventsForTemplate[$day]['display_date'];
      $eventDay = substr($eventDisplayDate, 0, 1);
      if ($eventDay === 'Sun') {
        continue;
      }
      if (count($eventsForTemplate[$day]['events']) > 0) {
        continue;
      }
      $eventsForTemplate[$day]['display_date'] = $eventDay;
    }
    // Add today flag for first and last dates in the course,
    // if it applies.
    $eventsForTemplate[1]['is_today'] = (1 === $todayAsDayNumber);
    $eventsForTemplate[$lastDay]['is_today'] = ($lastDay === $todayAsDayNumber);
    return [$problemMessage, $eventsForTemplate];
  }

  /**
   * Format a date for the timeline. E.g.: W Jul 21.
   *
   * @param \DateTime $when
   *   Date to format.
   *
   * @return string
   *   Result.
   */
  protected function formatTimelineDate(DateTime $when) {
    $longDoW = $when->format('D');
    $doWLetter = substr($longDoW, 0, 1);
    $result = $doWLetter . ' ' . $when->format(' M j');
    return $result;
  }

}
